from sqlalchemy import (
    CHAR,
    Column,
    DateTime,
    ForeignKey,
    MetaData,
    String,
    Table,
    func,
)
from sqlalchemy.dialects.oracle import BINARY_FLOAT, NUMBER
from sqlalchemy.orm import DeclarativeBase, column_property, relationship

from .enum_type import EnumType
from .enums import CodeEnum

OLDW_METADATA = MetaData(schema="OLDW")

course_subject = Table(
    "CRSE_SBJCT",
    OLDW_METADATA,
    Column("CRSE_KEY", ForeignKey("CRSE.CRSE_KEY"), primary_key=True),
    Column("SBJCT_KEY", ForeignKey("SBJCT.SBJCT_KEY"), primary_key=True),
)

course_bundle = Table(
    "CRSE_BUNDLE",
    OLDW_METADATA,
    Column("CRSE_KEY", ForeignKey("CRSE.CRSE_KEY"), primary_key=True),
    Column("BUNDLE_KEY", ForeignKey("BUNDLE.BUNDLE_KEY"), primary_key=True),
)


bundle_subject = Table(
    "BUNDLE_SBJCT",
    OLDW_METADATA,
    Column("BUNDLE_KEY", ForeignKey("BUNDLE.BUNDLE_KEY"), primary_key=True),
    Column("SBJCT_KEY", ForeignKey("SBJCT.SBJCT_KEY"), primary_key=True),
)

assessment_item = Table(
    "ASSESSMENT_ITEM",
    OLDW_METADATA,
    Column("ASSESSMENT_KEY", ForeignKey("ASSESSMENT.ASSESSMENT_KEY"), primary_key=True),
    Column("ITEM_KEY", ForeignKey("ITEM.ITEM_KEY"), primary_key=True),
)


def inst_repr(typ, attrs):
    params = [f"{key}='{value}'" for key, value in attrs.items() if value is not None]
    return f"<{typ}({', '.join(params)})>"


class Base(DeclarativeBase):
    metadata = OLDW_METADATA
    __repr_fields__ = ["id"]

    def __repr__(self):
        return inst_repr(
            self.__class__.__name__,
            {key: getattr(self, key) for key in self.__repr_fields__},
        )


class SourceSystem(Base):
    __tablename__ = "SRC_SYS"
    __repr_fields__ = ["id", "platform_name", "description"]

    id = Column("SRC_SYS_CD", String(50), primary_key=True)
    description = Column("SRC_SYS_DES", String(100), nullable=False)
    platform_name = Column("LRNG_PLTFRM_NM", String(100), nullable=False)
    load_date = Column("LOAD_DT", DateTime, nullable=True)

    courses = relationship("Course", back_populates="source_system")
    users = relationship("OnlineLearningUser", back_populates="source_system")
    subjects = relationship("Subject", back_populates="source_system")
    bundles = relationship("Bundle", back_populates="source_system")
    assessments = relationship("Assessment", back_populates="source_system")
    questions = relationship("Question", back_populates="source_system")


class Course(Base):
    __tablename__ = "CRSE"
    __repr_fields__ = ["id", "name", "slug", "source_system_id"]

    class LevelCode(CodeEnum):
        BEGINNER = "BEGINNER", "Beginner"
        INTERMEDIATE = "INTERMEDIATE", "Intermediate"
        ADVANCED = "ADVANCED", "Advanced"

    id = Column("CRSE_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_course_id = Column("SRC_SYS_CRSE_CD", String(255))
    name = Column("CRSE_NM", String(2000))
    slug = Column("CRSE_SLUG_TXT", String(4000), nullable=True)
    launched = Column("CRSE_LAUNCH_DTTM", DateTime, nullable=True)
    image_url = Column("CRSE_IMAGE_URL_TXT", String(4000), nullable=True)
    video_url = Column("CRSE_VIDEO_URL_TXT", String(4000), nullable=True)
    description = Column("CRSE_DES", String(4000), nullable=True)
    html_description = Column("CRSE_HTML_DES", String(4000), nullable=True)
    level_code = Column("CRSE_LVL_CD", EnumType(LevelCode, String(50)), nullable=True)
    res_ind = Column("CRSE_RES_IND", String(1), nullable=True)

    source_system = relationship(SourceSystem, back_populates="courses")
    classes = relationship("Class", back_populates="course")
    ratings = relationship("CourseRating", back_populates="course")
    course_instances = relationship("CourseInstance", back_populates="course")
    subjects = relationship(
        "Subject", secondary=course_subject, back_populates="courses"
    )
    bundles = relationship("Bundle", secondary=course_bundle, back_populates="courses")
    course_memberships = relationship("CourseMembership", back_populates="course")


class Class(Base):
    __tablename__ = "CLASS"
    __repr_fields__ = ["id", "course_id", "source_system_id"]

    id = Column("CLASS_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_class_id = Column("SRC_SYS_CLASS_CD", String(255))
    course_id = Column("CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY"))
    begin_date = Column("CLASS_BEGIN_DTTM", DateTime)
    end_date = Column("CLASS_END_DTTM", DateTime, nullable=True)
    enrollment_begin_date = Column("CLASS_ENRL_PRD_BEGIN_DTTM", DateTime, nullable=True)
    enrollment_end_date = Column("CLASS_ENRL_PRD_END_DTTM", DateTime, nullable=True)
    course_instance_id = Column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    is_dummy_class = Column("DUMMY_CLASS_IND", NUMBER(1), nullable=True)

    course_instance = relationship("CourseInstance", back_populates="classes")
    course = relationship("Course", back_populates="classes")
    class_memberships = relationship("ClassMembership", back_populates="class_")
    modules = relationship("Module", back_populates="class_")
    submodules = relationship("Submodule", back_populates="class_")
    items = relationship("Item", back_populates="class_")


class OnlineLearningUser(Base):
    __tablename__ = "ONLINE_LRNG_USR"
    __repr_fields__ = ["id", "full_name", "source_system_id", "source_system_user_id"]

    id = Column("USR_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_user_id = Column("SRC_SYS_USR_CD", String(50))
    last_login = Column("SRC_SYS_LAST_LOGIN_DTTM", DateTime, nullable=True)
    date_joined = Column("LRNG_PLTFRM_JOIN_DTTM", DateTime, nullable=True)
    first_name = Column("FIRST_NM", String(100), nullable=True)
    last_name = Column("LAST_NM", String(100), nullable=True)
    full_name = Column("FULL_NM", String(200), nullable=True)

    source_system = relationship(SourceSystem, back_populates="users")
    class_memberships = relationship("ClassMembership", back_populates="user")
    email_addresses = relationship("EmailAddress", back_populates="user")
    course_ratings = relationship("CourseRating", back_populates="user")
    bundle_memberships = relationship("BundleMembership", back_populates="user")
    course_memberships = relationship("CourseMembership", back_populates="user")


class RoleCode(CodeEnum):
    LEARNER = "LEARNER", "Learner"
    RESEARCHER = "RESEARCHER", "Researcher"
    ADMINISTRATOR = "ADMINISTRATOR", "Administrator"
    BROWSER = "BROWSER", "Browser"
    TEACHING_STAFF = "TEACHING_STAFF", "Teaching Staff"
    COURSE_ASSISTANT = "COURSE_ASSISTANT", "Course Assistant"
    UNENROLLED_LEARNER = "UNENROLLED_LEARNER", "Unenrolled Learner"
    PREENROLLED_LEARNER = "PRE_ENROLLED_LEARNER", "Pre-enrolled Learner"
    INSTRUCTOR = "INSTRUCTOR", "Instructor"
    MENTOR = "MENTOR", "Mentor"


class CertificateType(CodeEnum):
    VERIFIED = "VERIFIED", "Verified"
    AUDIT = "AUDIT", "Audit"
    HONORS = "HONOR", "Honors"


class PaymentSourceCode(CodeEnum):
    FINANCIAL_AID_AWARD = "FINANCIAL AID AWARD", "Financial Aid Award"
    ENTERPRISE = "ENTERPRISE", "Enterprise"
    INDIVIDUAL_PAYMENT = "INDIVIDUAL PAYMENT", "Individual Payment"


class ClassMembership(Base):
    __tablename__ = "CLASS_MBRSHP"
    __repr_fields__ = ["id", "role_code", "class_id", "user_id"]

    id = Column("CLASS_MBRSHP_KEY", CHAR(36), primary_key=True)
    class_id = Column("CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY"))
    user_id = Column("USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"))
    role_code = Column("MBRSHP_ROLE_CD", EnumType(RoleCode, String(50)))
    begin_date = Column("MBRSHP_BEGIN_DTTM", DateTime)
    eligible_certificate_type = Column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type = Column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date = Column("CERT_AWARD_DTTM", DateTime, nullable=True)
    payment_source_code = Column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )

    class_ = relationship(Class, back_populates="class_memberships")
    user = relationship(OnlineLearningUser, back_populates="class_memberships")


class EmailAddress(Base):
    __tablename__ = "USR_EMAIL_ADDR"
    __repr_fields__ = ["user_id", "email", "primary_address"]

    user_id = Column(
        "USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"), primary_key=True
    )
    email = Column("USR_EMAIL_ADDR_TXT", String(500), primary_key=True)
    primary_address = Column("PRMRY_EMAIL_ADDR_IND", NUMBER(1))

    user = relationship(OnlineLearningUser, back_populates="email_addresses")

    username = column_property(
        func.substr(email, 0, func.instr(email, "@", 1) - 1), deferred=True
    )
    domain = column_property(
        func.substr(email, func.instr(email, "@", 1) + 1), deferred=True
    )


class CourseRating(Base):
    __tablename__ = "USR_CRSE_RT"
    __repr_fields__ = ["id", "user_id", "course_id", "rating", "max_rating"]

    class RatingSystemCode(CodeEnum):
        STAR = "STAR", "Star"
        NPS_FIRST_WEEK = "NPS_FIRST_WEEK", "Net Promoter Score, First Week"
        NPS_END_OF_COURSE = "NPS_END_OF_COURSE", "Net Promoter Score, End of Course"

    id = Column("USR_CRSE_RT_KEY", CHAR(36), primary_key=True)
    user_id = Column("USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"))
    course_id = Column("CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY"))
    date = Column("CRSE_RT_DTTM", DateTime)
    rating_code = Column(
        "RT_SYS_CD", EnumType(RatingSystemCode, String(100)), nullable=True
    )
    rating = Column("CRSE_RT_NBR", NUMBER(38))
    max_rating = Column("CRSE_MAX_RT_NBR", NUMBER(38))
    comment = Column("CRSE_RT_CMNT_TXT", String(4000), nullable=True)

    user = relationship(OnlineLearningUser, back_populates="course_ratings")
    course = relationship(Course, back_populates="ratings")


class Subject(Base):
    __tablename__ = "SBJCT"
    __repr_fields__ = ["id", "name", "slug", "source_system_id"]

    id = Column("SBJCT_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_subject_id = Column("SRC_SYS_SBJCT_CD", String(255))
    html_description = Column("SBJCT_HTML_DES", String(4000))
    name = Column("SBJCT_NM", String(100))
    slug = Column("SBJCT_SLUG_TXT", String(4000), nullable=True)
    image_url = Column("SBJCT_IMAGE_URL_TXT", String(4000), nullable=True)
    um_subject_name = Column("UM_SBJCT_NM", String(100))

    source_system = relationship(SourceSystem, back_populates="subjects")
    courses = relationship(
        "Course", secondary=course_subject, back_populates="subjects"
    )
    bundles = relationship(
        "Bundle", secondary=bundle_subject, back_populates="subjects"
    )


class Bundle(Base):
    __tablename__ = "BUNDLE"
    __repr_fields__ = ["id", "name", "type", "slug", "source_system_id"]

    id = Column("BUNDLE_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_bundle_id = Column("SRC_SYS_BUNDLE_CD", String(255))
    source_system_bundle_instance_id = Column("SRC_SYS_BUNDLE_INSTC_CD", String(255))
    type = Column("BUNDLE_TYP_CD", String(50))
    name = Column("BUNDLE_NM", String(500))
    description = Column("BUNDLE_DES", String(4000), nullable=True)
    html_description = Column("BUNDLE_HTML_DES", String(4000), nullable=True)
    slug = Column("BUNDLE_SLUG_TXT", String(4000), nullable=True)
    version = Column("BUNDLE_VER_NBR", NUMBER(38), nullable=True)

    source_system = relationship(SourceSystem, back_populates="bundles")
    courses = relationship("Course", secondary=course_bundle, back_populates="bundles")
    subjects = relationship(
        "Subject", secondary=bundle_subject, back_populates="bundles"
    )
    bundle_memberships = relationship("BundleMembership", back_populates="bundle")

    # TODO: Define TypeCode enum for Bundle.type values


class CourseInstance(Base):
    __tablename__ = "CRSE_INSTC"
    __repr_fields__ = ["id", "course_id", "source_system_course_instance_id"]

    id = Column("CRSE_INSTC_KEY", CHAR(36), primary_key=True)
    course_id = Column("CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY"))
    source_system_course_instance_id = Column("SRC_SYS_CRSE_INSTC_CD", String(50))
    created_date = Column("CRSE_INSTC_CRT_DTTM", DateTime, nullable=True)

    classes = relationship(Class, back_populates="course_instance")
    course = relationship(Course, back_populates="course_instances")
    modules = relationship("Module", back_populates="course_instance")
    submodules = relationship("Submodule", back_populates="course_instance")
    items = relationship("Item", back_populates="course_instance")


class Module(Base):
    __tablename__ = "MODULE"
    __repr_fields__ = [
        "id",
        "course_instance_id",
        "class_id",
        "source_system_module_id",
    ]

    id = Column("MODULE_KEY", CHAR(36), primary_key=True)
    course_instance_id = Column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id = Column("CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY"))
    source_system_module_id = Column("SRC_SYS_MODULE_CD", String(500))
    order = Column("MODULE_ORDER_NBR", NUMBER(38), nullable=True)
    name = Column("MODULE_NM", String(4000), nullable=True)
    description = Column("MODULE_DES", String(4000), nullable=True)
    start_date = Column("MODULE_START_DTTM", DateTime, nullable=True)
    staff_access = Column("STAFF_ACCSS_IND", NUMBER(1), nullable=True)

    course_instance = relationship(CourseInstance, back_populates="modules")
    class_ = relationship(Class, back_populates="modules")
    submodules = relationship("Submodule", back_populates="module")


class Submodule(Base):
    __tablename__ = "SUBMODULE"
    __repr_fields__ = [
        "id",
        "course_instance_id",
        "class_id",
        "source_system_submodule_id",
    ]

    id = Column("SUBMODULE_KEY", CHAR(36), primary_key=True)
    course_instance_id = Column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id = Column("CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY"))
    source_system_submodule_id = Column("SRC_SYS_SUBMODULE_CD", String(500))
    parent_submodule_id = Column(
        "PRNT_SUBMODULE_KEY", CHAR(36), ForeignKey("SUBMODULE.SUBMODULE_KEY")
    )
    module_id = Column("MODULE_KEY", CHAR(36), ForeignKey("MODULE.MODULE_KEY"))
    order = Column("SUBMODULE_ORDER_NBR", NUMBER(38), nullable=True)
    name = Column("SUBMODULE_NM", String(4000), nullable=True)
    staff_access = Column("STAFF_ACCSS_IND", NUMBER(1), nullable=True)

    course_instance = relationship(CourseInstance, back_populates="submodules")
    class_ = relationship(Class, back_populates="submodules")
    module = relationship(Module, back_populates="submodules")
    parent = relationship("Submodule", backref="children", remote_side=[id])
    items = relationship("Item", back_populates="submodule")


class Item(Base):
    __tablename__ = "ITEM"
    __repr_fields__ = ["id", "course_instance_id", "source_system_item_id"]

    id = Column("ITEM_KEY", CHAR(36), primary_key=True)
    course_instance_id = Column(
        "CRSE_INSTC_KEY", CHAR(36), ForeignKey("CRSE_INSTC.CRSE_INSTC_KEY")
    )
    class_id = Column("CLASS_KEY", CHAR(36), ForeignKey("CLASS.CLASS_KEY"))
    source_system_item_id = Column("SRC_SYS_ITEM_CD", String(500))
    submodule_id = Column(
        "SUBMODULE_KEY", CHAR(36), ForeignKey("SUBMODULE.SUBMODULE_KEY")
    )
    order = Column("ITEM_ORDER_NBR", NUMBER(38), nullable=True)
    name = Column("ITEM_NM", String(255), nullable=True)
    type_description = Column("ITEM_TYP_DES", String(255), nullable=True)
    type_category_name = Column("ITEM_TYP_CTGRY_NM", String(255), nullable=True)
    is_graded = Column("ITEM_TYP_GRDD_IND", NUMBER(1), nullable=True)
    url = Column("ITEM_URL_TXT", String(4000), nullable=True)

    course_instance = relationship(CourseInstance, back_populates="items")
    class_ = relationship(Class, back_populates="items")
    submodule = relationship(Submodule, back_populates="items")
    assessments = relationship(
        "Assessment", secondary=assessment_item, back_populates="items"
    )


class Assessment(Base):
    __tablename__ = "ASSESSMENT"
    __repr_fields__ = ["id", "source_system_assessment_id", "category"]

    id = Column("ASSESSMENT_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_assessment_id = Column("SRC_SYS_ASSESSMENT_CD", String(500))
    category = Column("ASSESSMENT_CTGRY_CD", String(25))
    type = Column("ASSESSMENT_TYP_CD", String(50), nullable=True)
    pass_percentage = Column("PASS_PCT", BINARY_FLOAT, nullable=True)
    updated_date = Column("UPDT_DTTM", DateTime, nullable=True)
    ttl = Column("ASSESSMENT_TTL_TXT", String(500), nullable=True)
    max_submissions = Column("MAX_SBMSSN_NBR", NUMBER(38), nullable=True)
    version = Column("VERSION_NBR", NUMBER(38), nullable=True)
    max_score = Column("MAX_SCR", BINARY_FLOAT, nullable=True)

    source_system = relationship(SourceSystem, back_populates="assessments")
    items = relationship(Item, secondary=assessment_item, back_populates="assessments")
    assessment_questions = relationship(
        "AssessmentQuestion", back_populates="assessment"
    )
    peer_assignment = relationship(
        "PeerAssignment", uselist=False, back_populates="assessment"
    )
    programming_assignment = relationship(
        "ProgrammingAssignment", uselist=False, back_populates="assessment"
    )
    quiz = relationship("Quiz", uselist=False, back_populates="assessment")


class PeerAssignment(Base):
    __tablename__ = "PEER_ASSGN"

    id = Column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    required_review_count = Column("RQRD_RVW_NBR", NUMBER(38), nullable=True)
    required_reviewer_for_score_count = Column(
        "RQRD_RVWR_FOR_SCR_NBR", NUMBER(38), nullable=True
    )
    required_wait_for_score = Column(
        "RQRD_WAIT_FOR_SCR_DUR_NBR", NUMBER(38), nullable=True
    )

    assessment = relationship(Assessment, back_populates="peer_assignment")


class ProgrammingAssignment(Base):
    __tablename__ = "PGMMNG_ASSGN"

    id = Column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    submission_type = Column("SBMSSN_TYP_CD", String(50), nullable=True)
    instructions = Column("INSTRN_TXT", String(4000), nullable=True)
    submission_builder_schema_type = Column(
        "SBMSSN_BLDR_SCHEMA_TYP_CD", String(50), nullable=True
    )
    submission_builder_schema = Column(
        "SBMSSN_BLDR_SCHEMA_TXT", String(4000), nullable=True
    )

    assessment = relationship(Assessment, back_populates="programming_assignment")


class Quiz(Base):
    __tablename__ = "QUIZ"

    id = Column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    feedback_config = Column("FDBCK_CONFIG_CD", String(50), nullable=True)

    assessment = relationship(Assessment, back_populates="quiz")


class Question(Base):
    __tablename__ = "QSTN"
    __repr_fields__ = ["id", "source_system_id", "type"]

    id = Column("QSTN_KEY", CHAR(36), primary_key=True)
    source_system_id = Column(
        "SRC_SYS_CD", String(50), ForeignKey("SRC_SYS.SRC_SYS_CD")
    )
    source_system_question_id = Column("SRC_SYS_QSTN_CD", String(500))
    type = Column("QSTN_TYP_CD", String(50), nullable=True)
    prompt = Column("PROMPT_TXT", String(4000), nullable=True)
    is_shuffled = Column("SHUFFLE_IND", NUMBER(1), nullable=True)
    allows_partial_scoring = Column("PRTL_SCRNG_IND", NUMBER(1), nullable=True)
    default_incorrect_feedback = Column(
        "DFLT_INCORRECT_FDBCK_TXT", String(4000), nullable=True
    )
    feedback_text = Column("FDBCK_TXT", String(4000), nullable=True)

    source_system = relationship(SourceSystem, back_populates="questions")
    assessment_questions = relationship("AssessmentQuestion", back_populates="question")


class AssessmentQuestion(Base):
    __tablename__ = "ASSESSMENT_QSTN"
    __repr_fields__ = ["id", "source_system_id", "type"]

    assessment_id = Column(
        "ASSESSMENT_KEY",
        CHAR(36),
        ForeignKey("ASSESSMENT.ASSESSMENT_KEY"),
        primary_key=True,
    )
    question_id = Column(
        "QSTN_KEY", CHAR(36), ForeignKey("QSTN.QSTN_KEY"), primary_key=True
    )
    cuepoint = Column("CUEPOINT_NBR", NUMBER(38), nullable=True)
    order = Column("ORDER_NBR", NUMBER(38), nullable=True)
    weight = Column("WT_NBR", NUMBER(38), nullable=True)
    max_attempts = Column("MAX_ATTMPT_NBR", NUMBER(38), nullable=True)
    show_answer = Column("SHOW_ANSWR_CD", String(100), nullable=True)
    has_group_access = Column("GRP_ACCSS_IND", NUMBER(1), nullable=True)

    assessment = relationship(Assessment, back_populates="assessment_questions")
    question = relationship(Question, back_populates="assessment_questions")


class BundleMembership(Base):
    __tablename__ = "BUNDLE_MBRSHP"
    __repr_fields__ = ["id", "bundle_id", "user_id"]

    id = Column("BUNDLE_MBRSHP_KEY", CHAR(36), primary_key=True)
    bundle_id = Column("BUNDLE_KEY", CHAR(36), ForeignKey("BUNDLE.BUNDLE_KEY"))
    user_id = Column("USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"))
    role_code = Column("MBRSHP_ROLE_CD", EnumType(RoleCode, String(50)))
    begin_date = Column("MBRSHP_BEGIN_DTTM", DateTime)
    eligible_certificate_type = Column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type = Column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date = Column("CERT_AWARD_DTTM", DateTime, nullable=True)
    payment_source_code = Column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )

    bundle = relationship(Bundle, back_populates="bundle_memberships")
    user = relationship(OnlineLearningUser, back_populates="bundle_memberships")


class CourseMembership(Base):
    __tablename__ = "CRSE_MBRSHP"
    __repr_fields__ = ["id", "course_id", "user_id"]

    id = Column("CRSE_MBRSHP_KEY", CHAR(36), primary_key=True)
    course_id = Column("CRSE_KEY", CHAR(36), ForeignKey("CRSE.CRSE_KEY"))
    user_id = Column("USR_KEY", CHAR(36), ForeignKey("ONLINE_LRNG_USR.USR_KEY"))
    role_code = Column("MBRSHP_ROLE_CD", EnumType(RoleCode, String(50)))
    begin_date = Column("MBRSHP_BEGIN_DTTM", DateTime)
    eligible_certificate_type = Column(
        "ELIG_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    awarded_certificate_type = Column(
        "AWARD_CERT_TYP_CD", EnumType(CertificateType, String(50)), nullable=True
    )
    cert_award_date = Column("CERT_AWARD_DTTM", DateTime, nullable=True)
    payment_source_code = Column(
        "PMT_SRC_CD", EnumType(PaymentSourceCode, String(50)), nullable=True
    )
    overall_grade = Column("GRD_PCT", NUMBER(12, 9), nullable=True)

    course = relationship(Course, back_populates="course_memberships")
    user = relationship(OnlineLearningUser, back_populates="course_memberships")
